// manager.cpp
#include <vector>
#include <algorithm>
#include <allegro.h>
#include <iostream>
#include <fstream>
#include <ostream>

using namespace std;

#include "common.h"
#include "classes.h"
#include "manager.h"
#include "playerinfo.h"

typedef vector<GameObject *>::const_iterator ObjIterator;
typedef vector<Particle *>::const_iterator ParticleIterator;
typedef vector<Rect>::const_iterator RectIterator;

static vector<GameObject *> gameobjects;
static vector<GameObject *> deadobjects;
static vector<GameObject *> newobjects;

// Optimisation; only store the ones near the view
static vector<GameObject *> nearobjects;

static vector<Particle *> particles;
static vector<Rect> dirtyrectangles;

static GameObject * trackingobject = 0;
GameObject * playership = 0;

Point viewportcentre;
int screenoffsetx, screenoffsety;

static float quantum;
static int framecounter;
static volatile bool timerready;

static void ScreenDump()
{
	BITMAP *subbm = create_sub_bitmap(screen,0,0,640,480);
	save_bitmap("dump.pcx", subbm, 0);
	destroy_bitmap(subbm);
}

GameObject * findnearest(float x, float y)
{
	float mindist = 1e9;
	GameObject * found = 0;
	for (ObjIterator it = nearobjects.begin(); it != nearobjects.end(); it++)
	{
		GameObject *obj = *it;
		float xofs = x - obj->position.x;
		float yofs = y - obj->position.y;
		float dist = sqrt(xofs * xofs + yofs * yofs);
		if (dist < mindist) {
			found = obj;
			mindist = dist;
		}
	}
	return found;
}

void addobject(GameObject *obj)
{
	Particle *p = dynamic_cast<Particle *> (obj);
	if (p) {
		particles.push_back(p);
	} else {
		newobjects.push_back(obj);
	}
}

void trackobject(GameObject *obj)
{
	if (dynamic_cast<Ship *> (obj)) {
		playership = obj;
	}
	trackingobject = obj;
}

static void deletelistobjects(vector<GameObject *> &list)
{
	for (ObjIterator it = list.begin(); it != list.end(); it++)
	{
		GameObject *obj = *it;
		delete obj;
	}
	list.clear();
}

void initmanager()
{
	viewportcentre.x = 640 / 2;
	viewportcentre.y = 480 / 2;
	quantum = 0;
	deletelistobjects(gameobjects);
	deletelistobjects(deadobjects);
	deletelistobjects(newobjects);
	// Delete the particles manually as the above function won't work.
	for (ParticleIterator it = particles.begin(); it != particles.end(); it++)
	{
		Particle *obj = *it;
		delete obj;
	}
	particles.clear();
	// Nearobjects containes copies only, so it's safe to just clear it.
	nearobjects.clear();
	playership = 0;
	trackingobject = 0;
}

static void timerfunc()
{
	timerready = true;
}

static void track()
{
	if (trackingobject)
	{
		// copy the background to the screen
		float ofsx = trackingobject->position.x - viewportcentre.x;
		float ofsy = trackingobject->position.y - viewportcentre.y;
		float lensquared = (ofsx * ofsx) + (ofsy * ofsy);
		if (lensquared > 49) {
			viewportcentre.x += (ofsx / 10);
			viewportcentre.y += (ofsy / 10);
		}
		if (trackingobject->dead) {
			trackingobject = 0;
		}
		screenoffsetx = ((int) viewportcentre.x) - (640 / 2);
		screenoffsety = ((int) viewportcentre.y) - (480 / 2);
		if (screenoffsetx < 0) screenoffsetx = 0;
		if (screenoffsety < 0) screenoffsety = 0;
		if ((screenoffsetx + 640) >backgroundWidth)
			screenoffsetx = backgroundWidth - 640;
		if ((screenoffsety + 480) >backgroundHeight)
			screenoffsety = backgroundHeight - 480;
	}
}

void refreshvscreen()
{
	blit(background,vscreen, 0, 0, 0, 0, backgroundWidth, backgroundHeight);
}

static void recalibrate()
{
	// Call this after we've been doing something else
	// So that we don't screw up the timer.
	remove_int(timerfunc);
	framecounter = 0;
	quantum = 0;
	timerready = false;
	install_int(timerfunc,1000);
}

void mainloop()
{
	bool finished = false;
	timerready = false;
	framecounter = 0;
	refreshvscreen();
	// Initially centre the view on the ship
	if (trackingobject) {
		viewportcentre.x = trackingobject->position.x;
		viewportcentre.y = trackingobject->position.y;
		track();
	}
	// Now let's go - don't install the int until just before we start
	// For maximum accuracy.
	float finishtimer = 4;
	recalibrate();
	do {
		framecounter++;
		if (timerready) {
			float frames = (float) framecounter;
			if (frames < 20) frames = 20;
			quantum = 1.0 / frames;
			// cout << "FPS:" << framecounter << " Quantum: " << quantum << endl;
			framecounter = 0;
			timerready = false;
		}
		if (quantum >0) tickall(quantum);
		track();
		drawall();
		// Escape pressed - currently just quits.
		if (key[KEY_ESC]) {
			finished = true;
		}
		// Level complete
		if (playerinfo.levelcomplete)
			finished = true;
		// Dead
		if (playership == 0)
			finished = true;
		if (key[KEY_D]) {
			ScreenDump();
			recalibrate();
		}
		if (finished) finishtimer -= quantum;
	} while (finishtimer>0);
	remove_int(timerfunc);
}

void addrect(int x,int y, int w, int h)
{
	Rect r;
	r.x = x;
	r.y = y;
	r.w = w;
	r.h = h;
	dirtyrectangles.push_back(r);
}

// Bring out yer dead, bring out yer dead
// "There you go"
// "I'm not dead!"
static bool isDead(GameObject *obj) {
	return obj->dead;
}

void removedeads()
{
	if (playership != 0) {
		if (playership->dead) {
			playership = 0;
		}
	}
	// Remove dead things.
	// Remove all the ones which were deleted last time.
	deletelistobjects(deadobjects);

        // Remove all the dead ones from last tick. put them in deadObjects
	for (ObjIterator it = gameobjects.begin(); it != gameobjects.end(); it++)
	{
		GameObject *obj = *it;
		if (obj->dead) {
			deadobjects.push_back(obj);
		}
	}
        // Same for particles
	for (ParticleIterator it = particles.begin(); it != particles.end(); it++)
	{
		Particle *obj = *it;
		if (obj->dead) {
			deadobjects.push_back(obj);
		}
	}
	// Remove the dead ones.
	vector<GameObject *>::iterator new_end =
		remove_if(gameobjects.begin(), gameobjects.end(),isDead);
	gameobjects.erase(
		new_end,
		gameobjects.end());
	// Remove dead particles.
	{
		vector<Particle *>::iterator new_p_end =
			remove_if(particles.begin(), particles.end(),isDead);
		particles.erase(
			new_p_end,
			particles.end());
	}
	// Now, add any new objects (not particles, which are added directly)
	for (ObjIterator it = newobjects.begin(); it != newobjects.end(); it++)
	{
		GameObject *obj = *it;
			gameobjects.push_back(obj);
	}
	newobjects.clear();
}

static void checkcollision(GameObject * obj1, GameObject *obj2)
{
	CollidesWithScenery *c1, *c2;
	c1 = dynamic_cast <CollidesWithScenery *> (obj1);
	c2 = dynamic_cast <CollidesWithScenery *> (obj2);
	if (c1 == 0) return;
	if (c2 == 0) return;
	float dist, dx, dy;
	dx = c1->position.x - c2->position.x;
	dy = c1->position.y - c2->position.y;
	dist = sqrt(dx*dx + dy*dy);
	if (dist > (c1->radius + c2->radius))
		return;
	// Figure out relative velocity.
	float rvx = c1->velocity.x - c2->velocity.x;
	float rvy = c1->velocity.y - c2->velocity.y;
	// Normalise the relative position
	dx = dx / dist;
	dy = dy / dist;
	// Now dot product the two together to work out the component
	float component = (dx *rvx) + (dy * rvy);
	// It should be negative. I hope. This is because the relative velocity
	if (component >0)
		return;
	// should be bringing the objects together, not apart
	// Multiply by component to give the impulse.
	dx = dx * component;
	dy = dy * component;
	// Share out the impulse unevenly depending on mass.
	// The heavier object gets less impulse.
	float c2share = (c1->mass / (c2->mass + c1->mass)) * 2;
	float c1share = (c2->mass / (c2->mass + c1->mass)) * 2;
	// Add it to c2
	c2->velocity.x += dx * c2share;
	c2->velocity.y += dy * c2share;
	// Subtract from c1
	c1->velocity.x -= dx * c1share;
	c1->velocity.y -= dy * c1share;
	// Now call the relevant methods.
	c1->collidewithother(c2, component);
	c2->collidewithother(c1, component);
}

static void collidedetect()
{
	for (ObjIterator it1 = nearobjects.begin(); it1 != nearobjects.end(); it1++) {
		ObjIterator it2 = it1;
		it2++;
		for (; it2 != nearobjects.end(); it2++)
		{
			GameObject *obj1 = *it1;
			GameObject *obj2 = *it2;
			checkcollision(obj1, obj2);
		}
	}
}

void tickall(float quantum)
{
	removedeads();
	ObjIterator objend = gameobjects.end();
	for (ObjIterator it = gameobjects.begin(); it != objend; it++)
	{
		GameObject *obj = *it;
		obj->tick(quantum);
	}
	// And particles
	for (ParticleIterator it = particles.begin(); it != particles.end(); it++)
	{
		Particle *obj = *it;
		obj->tick(quantum);
	}
	collidedetect();
}

void drawall()
{
	nearobjects.clear();
	// Draw the objects
	for (ObjIterator it = gameobjects.begin(); it != gameobjects.end(); it++)
	{
		GameObject *obj = *it;
		// Cull objects too far away.
		float xdist = fabs(viewportcentre.x - obj->position.x);
		float ydist = fabs(viewportcentre.y - obj->position.y);
		if ((xdist < 640) && (ydist < 480)) {
			obj->draw();
			nearobjects.push_back(obj);
		}
	}
	// And particles
	for (ParticleIterator it = particles.begin(); it != particles.end(); it++)
	{
		Particle *obj = *it;
		obj->draw();
	}
	if (playerinfo.levelcomplete)
	{
		textout_centre(vscreen, font,
			"Level complete!", screenoffsetx + 320, screenoffsety+8, makecol(255,255,255));
		addrect(screenoffsetx, screenoffsety, 640, 16);
	}
	// copy the background to the screen
	blit(vscreen,screen, screenoffsetx, screenoffsety, 0, 0, 640, 480);
	// using dirty rectangles, blank out the areas modified.
	// erase dirty list.
	for (RectIterator it = dirtyrectangles.begin(); it != dirtyrectangles.end(); it++) {
		Rect r = *it;
		blit(background,vscreen, r.x, r.y, r.x, r.y, r.w, r.h);
	}
	dirtyrectangles.clear();
}

void saveobjects(char *filename)
{
	ofstream out(filename);
	if (out.bad()) {
		throw "Save error";
	}
	for (ObjIterator it = gameobjects.begin(); it != gameobjects.end(); it++)
	{
		GameObject *obj = *it;
		obj->save(out);
	}
}
